# Copyright 2020 AstroLab Software
# Author: Julien Peloton
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import json
import requests
from collections import OrderedDict
import numpy as np
import pandas as pd
import glob
import io
import zipfile

# Fink groupid
reporting_group_id = "103"

# ZTF data source
discovery_data_source_id = "48"

# instrument
instrument = "196"

# units
inst_units = "1"

# at type
at_type = "1"

# filters: 110 (g-ZTF), 111 (r-ZTF), 112 (i-ZTF)
filters_dict = {1: "110", 2: "111", 3: "112"}

reporter = "Julien Peloton, Anais Moller, Emille E. O. Ishida on behalf of the Fink broker"

remarks = "Early SN Ia candidate classified by Fink using the public ZTF stream. Object data at https://fink-portal.org/{} "

def search_tns(api_key, tns_marker, oid):
    """ Get TNS data for a given object ID

    Parameters
    ----------
    api_key: str
        API key for TNS
    tns_marker: str
        New marker to be inserted in the header (user-agent).
        See https://www.wis-tns.org/content/tns-newsfeed#comment-wrapper-23710
    oid: str
        Internal Object ID (e.g. ZTFXXXXXX)
    """
    orddict = OrderedDict(
        [
            ("units", "deg"),
            ("objname", ""),
            ("internal_name", oid)
        ]

    )

    search_data = [
        ('api_key', (None, api_key)),
        ('data', (None, json.dumps(orddict)))
    ]

    # define header
    headers = {'User-Agent': tns_marker}

    response = requests.post(
        "https://www.wis-tns.org/api/get/search",
        files=search_data,
        headers=headers,
        timeout=(5, 10)
    )

    try:
        reply = response.json()["data"]["reply"]
    except KeyError as e:
        reply = []

    return reply

def retrieve_groupid(api_key, tns_marker, oid):
    """ Get TNS groupid for a given object ID

    Parameters
    ----------
    api_key: str
        API key for TNS
    tns_marker: str
        New marker to be inserted in the header (user-agent).
        See https://www.wis-tns.org/content/tns-newsfeed#comment-wrapper-23710
    oid: str
        Internal Object ID (e.g. ZTFXXXXXX)
    """
    reply = search_tns(api_key, tns_marker, oid)

    if reply != []:
        objname = reply[0]["objname"]
    else:
        return -999

    data = {
        "objname": objname,
    }

    # get object type
    json_data = [
        ('api_key', (None, api_key)),
        ('data', (None, json.dumps(data)))
    ]

    # define header
    headers = {'User-Agent': tns_marker}

    response = requests.post(
        "https://www.wis-tns.org/api/get/object",
        files=json_data,
        headers=headers
    )

    data = response.json()['data']

    return data['reply']['discovery_data_source']['groupid']

def extract_radec(data):
    """ Return mean RA/Dec and scatter for a ZTF object based on all alerts

    Parameters
    ----------
    data: dict
        Dictionnary containing all alerts for a given ZTF objects

    Returns
    ----------
    out: dict
        Mean RA/Dec and scatter
    """
    ra = []
    dec = []

    ra.append(data['candidate']['ra'])
    for alert in data['prv_candidates']:
        ra.append(alert['ra'])
    ra = np.array(ra)

    dec.append(data['candidate']['dec'])
    for alert in data['prv_candidates']:
        dec.append(alert['dec'])
    dec = np.array(dec)

    mask = (ra != None) & (dec != None)
    return {
        'ra': np.mean(ra[mask]),
        'ra_err': np.std(ra[mask]),
        'dec': np.mean(dec[mask]),
        'dec_err': np.std(dec[mask])
    }

def read_past_ids(folder):
    """ Read all ZTF objectId already reported by Fink to TNS

    This is to avoid reporting twice the same objects

    Parameters
    ----------
    folder: str
        Path to the folder containing CSV files containing all ZTF objectId sent
    """
    filenames = glob.glob('{}/*.csv'.format(folder))
    if len(filenames) == 0:
        # no files found
        return pd.DataFrame()

    pdf = pd.concat(
        [
            pd.read_csv(i) for i in glob.glob('{}/*.csv'.format(folder))
        ]
    )
    return pdf

def download_catalog(api_key, tns_marker):
    """ Download entire TNS data (compressed csv file) into Pandas DataFrame

    Parameters
    ----------
    api_key: str
        Path to API key
    tns_marker: str
        New marker to be inserted in the header (user-agent).
        See https://www.wis-tns.org/content/tns-newsfeed#comment-wrapper-23710

    Returns
    ----------
    pdf_tns: Pandas DataFrame
        Pandas DataFrame with all the data
    """
    with open(api_key) as f:
        # remove line break...
        key = f.read().replace('\n', '')

    json_data = [
        ('api_key', (None, key)),
    ]

    # define header
    headers = {'User-Agent': tns_marker}

    r = requests.post(
      'https://www.wis-tns.org/system/files/tns_public_objects/tns_public_objects.csv.zip',
      files=json_data,
      headers=headers
    )

    with zipfile.ZipFile(io.BytesIO(r.content)) as myzip:
        data = myzip.read(name='tns_public_objects.csv')

    pdf_tns = pd.read_csv(io.BytesIO(data), skiprows=[0])

    return pdf_tns

def extract_ztf_entries(x):
    """Filter a string containg substrings starting with ZTF"""
    if x is None:
        return ""
    else:
        tmp = x.split(",")
        mask = [i.startswith("ZTF") for i in tmp]
        if sum(mask) > 0:
            return np.array(tmp)[mask][0]
        else:
            return ""

def print_fink_statistics(tns_logs_folder=None, tns_catalog=None):
    """Print Fink performance on TNS 

    Parameters
    ----------
    tns_logs_folder: str
        TNS log folder if available (only for operations). 
        Otherwise, if None, API call is made.
    tns_catalog: str
        Path to the TNS catalog in parquet.
    """
    # This could be replace by API call for Early SN Ia
    if tns_logs_folder is not None:
        tns_id = read_past_ids(tns_logs_folder)
    else:
        # Get latests 5 Early SN Ia candidates
        r = requests.post(
          "https://api.fink-portal.org/api/v1/latests",
          json={
            "class": "Early SN Ia candidate",
            "columns": "i:objectId",
            "n": "100000"
          }
        )

        # Format output in a DataFrame
        tmp = pd.read_json(io.BytesIO(r.content))
        tns_id = tmp.drop_duplicates("i:objectId").rename(columns={"i:objectId": "id"})
    cat = pd.read_parquet(tns_catalog)

    cat["id"] = cat["internalname"].apply(extract_ztf_entries)

    merged = pd.merge(tns_id, cat, on='id')

    print("Number of candidates sent to TNS: {}".format(len(tns_id)))
    print("Number of classified candidates: {}".format(len(merged)))
    print("--------------")
    print(merged.groupby("type").count().sort_values("id", ascending=False)["id"])




# Copyright 2020 AstroLab Software
# Author: Julien Peloton
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
import json
import requests
from collections import OrderedDict

from fink_tns.utils import extract_radec
from fink_tns.utils import inst_units, filters_dict, instrument
from fink_tns.utils import reporting_group_id, at_type, discovery_data_source_id
from fink_tns.utils import reporter, remarks

from astropy.time import Time
import pandas as pd
import numpy as np

def generate_photometry(data: dict):
    """ Define structure of the dictionnary that contains first detection data

    Parameters
    ----------
    data: dict
        Fink/ZTF alert

    Returns
    ---------
    dd: dict
    """
    dd = {
        "obsdate": "{}".format(Time(data['jd'], format='jd').fits.replace("T", " ")),
        "flux": "{}".format(data['magpsf']),
        "flux_error": "{}".format(data['sigmapsf']),
        "limiting_flux": "{}".format(data['diffmaglim']),
        "flux_units": "{}".format(inst_units),
        "filter_value": "{}".format(filters_dict[data['fid']]),
        "instrument_value": "{}".format(instrument),
        "exptime": "30",
        "observer": "Robot",
        "comments": "Data provided by ZTF, classified by Fink"
    }

    return dd

def generate_non_detection(data):
    """ Define structure of the dictionnary that contains last non-detection data

    Parameters
    ----------
    data: dict
        Fink/ZTF alert

    Returns
    ---------
    dd: dict
    """
    dd = {
        "obsdate": "{}".format(Time(data['jd'], format='jd').fits.replace("T", " ")),
        "limiting_flux": "{}".format(data['diffmaglim']),
        "flux_units": "{}".format(inst_units),
        "filter_value": "{}".format(filters_dict[data['fid']]),
        "instrument_value": "{}".format(instrument),
        "exptime": "30",
        "observer": "Robot",
        "comments": "Data provided by ZTF, classified by Fink"
    }

    return dd

def extract_discovery_photometry(data: dict) -> (dict, dict):
    """ Extract the photometry at the moment of discovery

    Parameters
    ----------
    data: dict
        Fink/ZTF alert data as a dictionnary

    Returns
    ----------
    first_photometry: dict
        Information about the first detection
    last_non_detection: dict
        Information about the last non-detection
    """
    tmp_pho = []
    tmp_upp = []

    # add candidate into photometry
    tmp_pho.append(generate_photometry(data['candidate']))

    # loop over prv_candidates, and add into photometry or non-det
    for alert in data['prv_candidates']:
        if alert['magpsf'] is not None:
            tmp_pho.append(generate_photometry(alert))
        else:
            tmp_upp.append(generate_non_detection(alert))

    # Sort photometry and keep the first one
    tmp_pho = sorted(tmp_pho, key=lambda i: i['obsdate'])
    first_photometry = tmp_pho[0]

    # Extract the last non-detection
    date_init = first_photometry['obsdate']
    filt_init = first_photometry['filter_value']
    tmp_upp = sorted(tmp_upp, key=lambda i: i['obsdate'])

    # Note we extract the last non-detection for the same filter!
    tmp_upp = [i for i in tmp_upp if (i['obsdate'] <= date_init) & (i['filter_value'] == filt_init)]

    if len(tmp_upp) == 0:
        last_non_detection = {
            "archiveid": "0",
            "archival_remarks": "ZTF non-detection limits not available"
        }
    else:
        last_non_detection = tmp_upp[-1]

    return first_photometry, last_non_detection

def extract_discovery_photometry_api(data: pd.DataFrame) -> (dict, dict):
    """ Extract the photometry at the moment of discovery

    Parameters
    ----------
    data: pd.DataFrame
        Fink/ZTF alert data from the REST API as a pandas DataFrame

    Returns
    ----------
    first_photometry: dict
        Information about the first detection
    last_non_detection: dict
        Information about the last non-detection
    """
    mask_valid = (data['d:tag'].values == 'valid') | (data['d:tag'].values == 'badquality')

    # first valid
    first = data[mask_valid].tail(1)

    mask_time = data[~mask_valid]['i:jd'] < first['i:jd'].values[0]

    # last non-detection
    last = data[~mask_valid][mask_time].head(1)

    first_photometry = {
        "obsdate": "{}".format(Time(first['i:jd'].values[0], format='jd').fits.replace("T", " ")),
        "flux": "{}".format(first['i:magpsf'].values[0]),
        "flux_error": "{}".format(first['i:sigmapsf'].values[0]),
        "limiting_flux": "{}".format(first['i:diffmaglim'].values[0]),
        "flux_units": "{}".format(inst_units),
        "filter_value": "{}".format(filters_dict[first['i:fid'].values[0]]),
        "instrument_value": "{}".format(instrument),
        "exptime": "30",
        "observer": "Robot",
        "comments": "Data provided by ZTF, classified by Fink"
    }

    if not last.empty:
        last_non_detection = {
            "obsdate": "{}".format(Time(last['i:jd'].values[0], format='jd').fits.replace("T", " ")),
            "limiting_flux": "{}".format(last['i:diffmaglim'].values[0]),
            "flux_units": "{}".format(inst_units),
            "filter_value": "{}".format(filters_dict[last['i:fid'].values[0]]),
            "instrument_value": "{}".format(instrument),
            "exptime": "30",
            "observer": "Robot",
            "comments": "Data provided by ZTF, classified by Fink"
        }
    else:
        last_non_detection = {}

    return first_photometry, last_non_detection

def build_report(
        data: dict, photometry: dict, non_detection: dict,
        reporter_custom=None, remarks_custom=None, at_type_=None) -> dict:
    """ Build json report to send to TNS

    Parameters
    ----------
    data: dict
        Mean RA/Dec and scatter
    photometry: dict
        Information about the first detection
    non_detection: dict
        Information about the last non-detection
    remarks_custom: str, optional
        Comments that will be displayed on TNS. Default is None, that is
        `utils.remarks`.
    reporter_custom: str, optional
        Names of the reporters that will be displayed on TNS.
        Default is None, that is `utils.reporter`.

    Returns
    ----------
    report: dict
        Dictionnary at the TNS format.
    """
    if remarks_custom is None:
        remarks_custom = remarks
    if reporter_custom is None:
        reporter_custom = reporter
    if at_type_ is None:
        at_type_ = at_type

    radec = extract_radec(data)
    report = {
        "ra": {
            "value": radec['ra'],
            "error": radec['ra_err'] * 3600,
            "units": "arcsec"
        },
        "dec": {
            "value": radec['dec'],
            "error": radec['dec_err'] * 3600,
            "units": "arcsec"
        },
        "reporting_group_id": reporting_group_id,
        "discovery_data_source_id": discovery_data_source_id,
        "reporter": reporter_custom,
        "discovery_datetime": photometry['obsdate'],
        "at_type": at_type_,
        "internal_name": data['objectId'],
        "remarks": remarks_custom.format(data['objectId']),
        "non_detection": non_detection,
        "photometry": {"photometry_group": {'0': photometry}}
    }

    return report

def build_report_api(
        data: pd.DataFrame, photometry: dict, non_detection: dict,
        reporter_custom=None, remarks_custom=None, at_type_=None) -> dict:
    """ Build json report to send to TNS

    Parameters
    ----------
    data: dict
        Data from the REST API /api/v1/objects
    photometry: dict
        Information about the first detection
    non_detection: dict
        Information about the last non-detection
    remarks_custom: str, optional
        Comments that will be displayed on TNS. Default is None, that is
        `utils.remarks`.
    reporter_custom: str, optional
        Names of the reporters that will be displayed on TNS.
        Default is None, that is `utils.reporter`.

    Returns
    ----------
    report: dict
        Dictionnary at the TNS format.
    """
    if remarks_custom is None:
        remarks_custom = remarks
    if reporter_custom is None:
        reporter_custom = reporter
    if at_type_ is None:
        at_type_ = at_type

    mask = ~np.isnan(data['i:ra'].values) & ~np.isnan(data['i:dec'].values)
    radec = {
        'ra': np.mean(data['i:ra'].values[mask]),
        'ra_err': np.std(data['i:ra'].values[mask]),
        'dec': np.mean(data['i:dec'].values[mask]),
        'dec_err': np.std(data['i:dec'].values[mask])
    }

    report = {
        "ra": {
            "value": radec['ra'],
            "error": radec['ra_err'] * 3600,
            "units": "arcsec"
        },
        "dec": {
            "value": radec['dec'],
            "error": radec['dec_err'] * 3600,
            "units": "arcsec"
        },
        "reporting_group_id": reporting_group_id,
        "discovery_data_source_id": discovery_data_source_id,
        "reporter": reporter_custom,
        "discovery_datetime": photometry['obsdate'],
        "at_type": at_type,
        "internal_name": data['i:objectId'].values[0],
        "remarks": remarks_custom.format(data['i:objectId'].values[0]),
        "non_detection": non_detection,
        "photometry": {"photometry_group": {'0': photometry}}
    }

    return report

def save_logs_and_return_json_report(name: str, folder: str, ids: list, report: dict):
    """ Save logs on disk (JSON), and return the path to the JSON file

    Parameters
    ----------
    name: str
        Filename
    folder: str
        Folder (or path) that will contain the log
    ids: list
        List of objectId that will be sent
    report: dict
        Payload to send to TNS

    Returns
    ---------
    json_report: str
        Path to the log
    """
    os.makedirs(folder, exist_ok=True)

    # Save processed ids on disk
    pdf_ids = pd.DataFrame.from_dict({'id': ids})
    pdf_ids.to_csv('{}/{}.csv'.format(folder, name), index=False)

    # Save report on disk
    json_report = '{}/{}.json'.format(folder, name)
    with open(json_report, 'w') as outfile:
        json.dump(report, outfile)
    return json_report

def format_to_json(source):
    """ change data to json format and return """
    parsed = json.loads(source, object_pairs_hook=OrderedDict)
    result = json.dumps(parsed, indent=4)
    return result

def send_json_report(api_key, url, json_file_path, tns_marker) -> int:
    """ Function for sending json reports (AT or Classification)

    Parameters
    ----------
    api_key: str
        API key for TNS
    url: str
        URL for posting results
    json_file_path: str
        Path to the generated report to be posted
    tns_marker: str
        New marker to be inserted in the header (user-agent).
        See https://www.wis-tns.org/content/tns-newsfeed#comment-wrapper-23710
    """
    # url for sending json reports
    json_url = url + '/set/bulk-report'

    # read json data from file
    json_read = format_to_json(open(json_file_path).read())

    # construct list of (key,value) pairs
    json_data = [
        ('api_key', (None, api_key)),
        ('data', (None, json_read))
    ]

    # define header
    headers = {'User-Agent': tns_marker}

    # send json report using request module
    response = requests.post(json_url, data=json_data, headers=headers)

    return response



# Copyright 2020 AstroLab Software
# Author: Julien Peloton
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import setuptools
from fink_tns import __version__

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="fink-tns",
    version=__version__,
    author="JulienPeloton",
    author_email="peloton@lal.in2p3.fr",
    description="Push Fink candidates to TNS.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: Unix",
        "Development Status :: 4 - Beta",
        "Intended Audience :: Science/Research",
    ],
    project_urls={
        'Documentation': "https://fink-broker.org",
        'Source': 'https://github.com/astrolabsoftware/fink-tns'
    },
)
